/* Copyright (C) 2001  James Niemira (urmane@urmane.org)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/* This is a Python extension to use Sendmail's libmilter functionality.  Put
   this file in <pythonsrc>/Modules and
   add this to <pythonsrc>/Modules/Setup:
   milter milter.c -I<libmilter includes> -L<libmilter lib location> -lmilter
   -L<libsmutil lib location> -lsmutil -lresolv
   Those lib locations are usually under the sendmail compile directory, eg,
   /usr/local/src/sendmail-8.11.1/obj.SunOS.5.6.sun4/libmilter on my box.
*/

#include <Python.h>
#include <libmilter/mfapi.h>

static PyObject *connect_callback = NULL;
static PyObject *helo_callback    = NULL;
static PyObject *envfrom_callback = NULL;
static PyObject *envrcpt_callback = NULL;
static PyObject *header_callback  = NULL;
static PyObject *eoh_callback     = NULL;
static PyObject *body_callback    = NULL;
static PyObject *eom_callback     = NULL;
static PyObject *abort_callback   = NULL;
static PyObject *close_callback   = NULL;

struct smfiDesc description; /* forward declaration */

/* This is a cheap kludge, but I think it works, because it's set individually
   in each callback wrapper function, which then calls the Python callback fn.
   Any function that needs the context from this side - ie, setreply, chgheader,
   etc - has to be called from within those Python functions.  The wrapper
   functions here use the value set earlier, which should be valid until the
   whole stack unwinds back thru the initial callback wrappers.  I think. */
static SMFICTX *context = NULL;

static PyObject *MilterError;

/* This is just a useful function.  I will probably ditch it in favor of
   some exceptions. */
static PyObject *
_generic_return(int val, char *errstr)
{
  if (val == MI_SUCCESS) {
    return Py_BuildValue("i", MI_SUCCESS);
  } else {
    PyErr_SetString(MilterError, errstr);
    /* return Py_BuildValue("i", MI_FAILURE); */
    return NULL;
  }
}

static char milter_set_name__doc__[] =
"set_name(String) -> None\n\
Set name of the current filter (as seen by Sendmail).";

static PyObject *
milter_set_name(PyObject *self, PyObject *args)
{
  char *name;

  if (!PyArg_ParseTuple(args, "s", &name))
    return NULL;
  description.xxfi_name = name;
  Py_INCREF(Py_None);
  return Py_None;
}

static char milter_set_flags__doc__[] =
"set_flags(int) -> None\n\
Set flags for filter capabilities; OR of one or more of:\n\
ADDHDRS - filter may add headers\n\
CHGBODY - filter may replace body\n\
ADDRCPT - filter may add recipients\n\
DELRCPT - filter may delete recipients\n\
CHGHDRS - filter may change/delete headers";

static PyObject *
milter_set_flags(PyObject *self, PyObject *args)
{
  int flags;

  if (!PyArg_ParseTuple(args, "i", &flags))
    return NULL;
  description.xxfi_flags = flags;
  Py_INCREF(Py_None);
  return Py_None;
}

static char milter_set_connect_callback__doc__[] =
"set_connect_callback(Function) -> None\n\
Sets the Python function invoked when a connection is made to sendmail.\n\
Function takes args (hostname, integer, hostaddr), where:\n\
hostname -> String - the connecting hostname\n\
integer -> int - the protocol family\n\
hostaddr -> String - the connecting host address\n\
(hostaddr may not be implemented correctly)";

static PyObject *
milter_set_connect_callback(PyObject *self, PyObject *args)
{
   PyObject *callback;

   if (!PyArg_ParseTuple(args, "O:set_connect_callback", &callback))
      return NULL;
   if (!PyCallable_Check(callback)) {
      PyErr_SetString(PyExc_TypeError, "parameter must be callable");
      return NULL;
   }
   Py_XINCREF(callback);
   Py_XDECREF(connect_callback);
   connect_callback = callback;
   Py_INCREF(Py_None);
   return Py_None;
}

static char milter_set_helo_callback__doc__[] =
"set_helo_callback(Function) -> None\n\
Sets the Python function invoked upon SMTP HELO.\n\
Function takes args (hostname), where:\n\
hostname -> String - the name given with the helo command.";

static PyObject *
milter_set_helo_callback(PyObject *self, PyObject *args)
{
   PyObject *callback;

   if (!PyArg_ParseTuple(args, "O:set_helo_callback", &callback))
      return NULL;
   if (!PyCallable_Check(callback)) {
      PyErr_SetString(PyExc_TypeError, "parameter must be callable");
      return NULL;
   }
   Py_XINCREF(callback);
   Py_XDECREF(helo_callback);
   helo_callback = callback;
   Py_INCREF(Py_None);
   return Py_None;
}

static char milter_set_envfrom_callback__doc__[] =
"set_envfrom_callback(Function) -> None\n\
Sets the Python function invoked on envelope from.\n\
Function takes args (*str), where\n\
str -> List of Strings - each string is (apparently) one source.\
(The api is a little unclear on why this isn't just one string)";

static PyObject *
milter_set_envfrom_callback(PyObject *self, PyObject *args)
{
   PyObject *callback;

   if (!PyArg_ParseTuple(args, "O:set_envfrom_callback", &callback))
      return NULL;
   if (!PyCallable_Check(callback)) {
      PyErr_SetString(PyExc_TypeError, "parameter must be callable");
      return NULL;
   }
   Py_XINCREF(callback);
   Py_XDECREF(envfrom_callback);
   envfrom_callback = callback;
   Py_INCREF(Py_None);
   return Py_None;
}

static char milter_set_envrcpt_callback__doc__[] =
"set_envrcpt_callback(Function) -> None\n\
Sets the Python function invoked on each envelope recipient.\n\
Function takes args (*str), where\n\
str -> List of Strings - each string is (apparently) one recipient.\n\
(The api is a little unclear on why this isn't just one string)";

static PyObject *
milter_set_envrcpt_callback(PyObject *self, PyObject *args)
{
   PyObject *callback;

   if (!PyArg_ParseTuple(args, "O:set_envrcpt_callback", &callback))
      return NULL;
   if (!PyCallable_Check(callback)) {
      PyErr_SetString(PyExc_TypeError, "parameter must be callable");
      return NULL;
   }
   Py_XINCREF(callback);
   Py_XDECREF(envrcpt_callback);
   envrcpt_callback = callback;
   Py_INCREF(Py_None);
   return Py_None;
}

static char milter_set_header_callback__doc__[] =
"set_header_callback(Function) -> None\n\
Sets the Python function invoked on each message header.\n\
Function takes args (field, value), where:\n\
field -> String - the header\n\
value -> String - the header's value";

static PyObject *
milter_set_header_callback(PyObject *self, PyObject *args)
{
   PyObject *callback;

   if (!PyArg_ParseTuple(args, "O:set_header_callback", &callback))
      return NULL;
   if (!PyCallable_Check(callback)) {
      PyErr_SetString(PyExc_TypeError, "parameter must be callable");
      return NULL;
   }
   Py_XINCREF(callback);
   Py_XDECREF(header_callback);
   header_callback = callback;
   Py_INCREF(Py_None);
   return Py_None;
}

static char milter_set_eoh_callback__doc__[] =
"set_eoh_callback(Function) -> None\n\
Sets the Python function invoked at end of header.\n\
Function takes no args.";

static PyObject *
milter_set_eoh_callback(PyObject *self, PyObject *args)
{
   PyObject *callback;

   if (!PyArg_ParseTuple(args, "O:set_eoh_callback", &callback))
      return NULL;
   if (!PyCallable_Check(callback)) {
      PyErr_SetString(PyExc_TypeError, "parameter must be callable");
      return NULL;
   }
   Py_XINCREF(callback);
   Py_XDECREF(eoh_callback);
   eoh_callback = callback;
   Py_INCREF(Py_None);
   return Py_None;
}

static char milter_set_body_callback__doc__[] =
"set_body_callback(Function) -> None\n\
Sets the Python function invoked for each body chunk. There may \
be multiple body chunks passed to the filter. End-of-lines are \
represented as received from SMTP (normally Carriage-Return/Line-Feed).\n\
Function takes args (chunk), where:\n\
chunk -> String - body data";

static PyObject *
milter_set_body_callback(PyObject *self, PyObject *args)
{
   PyObject *callback;

   if (!PyArg_ParseTuple(args, "O:set_body_callback", &callback))
      return NULL;
   if (!PyCallable_Check(callback)) {
      PyErr_SetString(PyExc_TypeError, "parameter must be callable");
      return NULL;
   }
   Py_XINCREF(callback);
   Py_XDECREF(body_callback);
   body_callback = callback;
   Py_INCREF(Py_None);
   return Py_None;
}

static char milter_set_eom_callback__doc__[] =
"set_eom_callback(Function) -> None\n\
Sets the Python function invoked at end of message. This routine is the only \
place where special operations such as modifying the message header, body, or \
envelope can be used.\n\
Function takes no args.";

static PyObject *
milter_set_eom_callback(PyObject *self, PyObject *args)
{
   PyObject *callback;

   if (!PyArg_ParseTuple(args, "O:set_eom_callback", &callback))
      return NULL;
   if (!PyCallable_Check(callback)) {
      PyErr_SetString(PyExc_TypeError, "parameter must be callable");
      return NULL;
   }
   Py_XINCREF(callback);
   Py_XDECREF(eom_callback);
   eom_callback = callback;
   Py_INCREF(Py_None);
   return Py_None;
}

static char milter_set_abort_callback__doc__[] =
"set_abort_callback(Function) -> None\n\
Sets the Python function invoked if message is aborted outside of the control \
of the filter, for example, if the SMTP sender issues an RSET command. If the \
abort callback is called, the eom callback will not be called and vice \
versa.\n\
Function takes no args.";

static PyObject *
milter_set_abort_callback(PyObject *self, PyObject *args)
{
   PyObject *callback;

   if (!PyArg_ParseTuple(args, "O:set_abort_callback", &callback))
      return NULL;
   if (!PyCallable_Check(callback)) {
      PyErr_SetString(PyExc_TypeError, "parameter must be callable");
      return NULL;
   }
   Py_XINCREF(callback);
   Py_XDECREF(abort_callback);
   abort_callback = callback;
   Py_INCREF(Py_None);
   return Py_None;
}

static char milter_set_close_callback__doc__[] =
"set_close_callback(Function) -> None\n\
Sets the Python function invoked at end of the connection. This is called on \
close even if the previous mail transaction was aborted.\n\
Function takes no args.";

static PyObject *
milter_set_close_callback(PyObject *self, PyObject *args)
{
   PyObject *callback;

   if (!PyArg_ParseTuple(args, "O:set_close_callback", &callback))
      return NULL;
   if (!PyCallable_Check(callback)) {
      PyErr_SetString(PyExc_TypeError, "parameter must be callable");
      return NULL;
   }
   Py_XINCREF(callback);
   Py_XDECREF(close_callback);
   close_callback = callback;
   Py_INCREF(Py_None);
   return Py_None;
}

static int
_generic_wrapper(PyObject *callback, PyObject *arglist)
{
  PyObject *result;
  PyObject *error;
  int retval;

  if (arglist == NULL)
    return SMFIS_CONTINUE; 
  result = PyEval_CallObject(callback, arglist);
  Py_DECREF(arglist);
  if (result == NULL) {
    if ((error = PyErr_Occurred())) {
      smfi_setreply(context, "452", "4.3.0", "Filter failure");
      PyErr_Clear();
      return SMFIS_TEMPFAIL;
    }
    /* Null but no error?  Erm, continue, I guess ... */
    return SMFIS_CONTINUE;
  }
  if (!PyArg_ParseTuple(result, "i", &retval)) {
    Py_DECREF(result);
    PyErr_Clear();
    return SMFIS_CONTINUE;
  }
  Py_DECREF(result);
  return retval;
}

int
milter_wrap_connect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *hostaddr)
{
  PyObject *arglist;
  
  if (connect_callback == NULL)
    return SMFIS_CONTINUE;
  context = ctx;
  arglist = Py_BuildValue("(shs)", hostname, hostaddr->sa_family,
			  hostaddr->sa_data);
  return _generic_wrapper(connect_callback, arglist);
}

int
milter_wrap_helo(SMFICTX *ctx, char *helohost)
{
   PyObject *arglist;

   if (helo_callback == NULL)
      return SMFIS_CONTINUE;
   context = ctx;
   arglist = Py_BuildValue("(s)", helohost);
   return _generic_wrapper(helo_callback, arglist);
}

int
milter_wrap_envfrom(SMFICTX *ctx, char **argv)
{
   PyObject *arglist;
   PyObject *o;
   int count;
   int i;
   char **p;
   char *buf;

   if (envfrom_callback == NULL)
      return SMFIS_CONTINUE; 
   context = ctx;
   /* Set initial vals */
   count = 0;
   p = argv;
   /* Count how many strings we've been passed */
   while (*p++ != NULL) count++;
   /* Allocate space for one 's' per string, two parens, and a \0 */
   buf = (char *)malloc(count + 3);
   if (buf == NULL)
     return SMFIS_CONTINUE;
   /* Construct the template string */
   buf[0] = '(';
   for (i=1;i<=count;i++)
     buf[i] = 's';
   buf[count+1] = ')';
   buf[count+2] = '\0';
   /* how to build the value in steps?  Cheat by copying from */
   /* Python/modsupport.c do_mktuple() and do_mkvalue() */
   if ((arglist = PyTuple_New(count)) == NULL) {
     free(buf);
     return SMFIS_CONTINUE;
   }
   for (i=0;i<count;i++) {
     /* There's some error checking performed in do_mkvalue() for a string */
     /* that's not currently done here - it probably should be */
     o = PyString_FromStringAndSize(argv[i], strlen(argv[i]));
     if (o == NULL) {
       Py_DECREF(o);
       free(buf);
       return SMFIS_CONTINUE;
     }
     PyTuple_SetItem(arglist, i, o);
   }
   return _generic_wrapper(envfrom_callback, arglist);
}

int
milter_wrap_envrcpt(SMFICTX *ctx, char **argv)
{
   PyObject *arglist;
   PyObject *o;
   int count;
   int i;
   char **p;
   char *buf;

   if (envrcpt_callback == NULL)              
      return SMFIS_CONTINUE;
   context = ctx;
   /* Set initial vals */
   count = 0;
   p = argv;
   /* Count how many strings we've been passed */
   while (*p++ != NULL) count++;
   /* Allocate space for one 's' per string, two parens, and a \0 */
   buf = (char *)malloc(count + 3);
   if (buf == NULL)
     return SMFIS_CONTINUE;
   /* Construct the template string */
   buf[0] = '(';
   for (i=1;i<=count;i++)
     buf[i] = 's';
   buf[count+1] = ')';
   buf[count+2] = '\0';
   /* how to build the value in steps?  Cheat by copying from */
   /* Python/modsupport.c do_mktuple() and do_mkvalue() */
   if ((arglist = PyTuple_New(count)) == NULL) {
     free(buf);
     return SMFIS_CONTINUE;
   }
   for (i=0;i<count;i++) {
     /* There's some error checking performed in do_mkvalue() for a string */
     /* that's not currently done here - it probably should be */
     o = PyString_FromStringAndSize(argv[i], strlen(argv[i]));
     if (o == NULL) {
       Py_DECREF(o);
       free(buf);
       return SMFIS_CONTINUE;
     }
     PyTuple_SetItem(arglist, i, o);
   }
   return _generic_wrapper(envrcpt_callback, arglist);
}    
  
int
milter_wrap_header(SMFICTX *ctx, char *headerf, char *headerv)
{
   PyObject *arglist;

   if (header_callback == NULL)
      return SMFIS_CONTINUE;
   context = ctx;
   arglist = Py_BuildValue("(ss)", headerf, headerv);
   return _generic_wrapper(header_callback, arglist);
}    

int
milter_wrap_eoh(SMFICTX *ctx)
{
   PyObject *arglist;

   if (eoh_callback == NULL)
      return SMFIS_CONTINUE;
   context = ctx;
   arglist = Py_BuildValue("()");
   return _generic_wrapper(eoh_callback, arglist);
}   

int
milter_wrap_body(SMFICTX *ctx, u_char *bodyp, size_t bodylen)
{
   PyObject *arglist;

   if (body_callback == NULL)
      return SMFIS_CONTINUE;
   context = ctx;
   /* Unclear whether this should be s#, z#, or t# */
   arglist = Py_BuildValue("(s#)", bodyp, bodylen);
   return _generic_wrapper(body_callback, arglist);
}  

int
milter_wrap_eom(SMFICTX *ctx)
{
   PyObject *arglist;

   if (eom_callback == NULL)
      return SMFIS_CONTINUE;
   context = ctx;
   arglist = Py_BuildValue("()");
   return _generic_wrapper(eom_callback, arglist);
}

int
milter_wrap_abort(SMFICTX *ctx)
{
   PyObject *arglist;

   if (abort_callback == NULL)
      return SMFIS_CONTINUE;
   context = ctx;
   arglist = Py_BuildValue("()");
   return _generic_wrapper(abort_callback, arglist);
}

int
milter_wrap_close(SMFICTX *ctx)
{
   PyObject *arglist;
                    
   if (close_callback == NULL)              
      return SMFIS_CONTINUE; 
   context = ctx;                           
   arglist = Py_BuildValue("()");                            
   return _generic_wrapper(close_callback, arglist);
}  

static char milter_register__doc__[] =
"register() -> SUCCESS or FAILURE\n\
Registers the current name, callbacks, and flags.  Required before main() \
is called.";

static PyObject *
milter_register(PyObject *self, PyObject *args)
{
   if (!PyArg_ParseTuple(args, ""))
      return NULL;
   return _generic_return(smfi_register(description), "cannot register");
}

static char milter_main__doc__[] =
"main() -> SUCCESS or FAILURE\n\
Main milter routine.  Set any callbacks, name, and flags desired, then call \
setconn(), then call register(), and finally call main().";

static PyObject *
milter_main(PyObject *self, PyObject *args)
{
   if (!PyArg_ParseTuple(args, ""))
      return NULL;
   return _generic_return(smfi_main(), "cannot run main");
}

static char milter_setdbg__doc__[] =
"setdbg(int) -> SUCCESS or FAILURE\n\
Sets debug level in sendmail/libmilter source.  Dubious usefulness.";

static PyObject *
milter_setdbg(PyObject *self, PyObject *args)
{
   int val;

   if (!PyArg_ParseTuple(args, "i", &val))
      return NULL;
   return _generic_return(smfi_setdbg(val), "cannot set debug value");
}

static char milter_settimeout__doc__[] =
"settimeout(int) -> SUCCESS or FAILURE\n\
Set the time (in seconds) that sendmail will wait before considering this \
filter dead.";

static PyObject *
milter_settimeout(PyObject *self, PyObject *args)
{
   int val;

   if (!PyArg_ParseTuple(args, "i", &val))
      return NULL;
   return _generic_return(smfi_settimeout(val), "cannot set timeout");
}

static char milter_setconn__doc__[] =
"setconn(filename) -> SUCCESS or FAILURE\n\
Sets the pathname to the unix socket that sendmail will use to communicate \
with this filter.  It must not exist, and sendmail will throw warnings if, \
eg, the file is under a group or world writable directory.  This call is \
mandatory, and is invoked before register() and main().";

static PyObject *
milter_setconn(PyObject *self, PyObject *args)
{
   char *str;

   if (!PyArg_ParseTuple(args, "s", &str))
      return NULL;
   return _generic_return(smfi_setconn(str), "cannot set connection");
}

static char milter_stop__doc__[] =
"stop() -> SUCCESS or FAILURE\n\
This function appears to be a controlled method to tell sendmail to\
stop using this filter.  It will close the socket.";

static PyObject *
milter_stop(PyObject *self, PyObject *args)
{
  PyObject *o;

  if (!PyArg_ParseTuple(args, ""))
    return NULL;
  /* This is to balance the incref done in setpriv.  The only place we know */
  /* that getpriv will no longer be called is after stop, so this is the */
  /* only place that's safe to decref again. */
  if ((context != NULL) && (o = (PyObject *)smfi_getpriv(context))) {
    Py_XDECREF(o);
  }
  return _generic_return(smfi_stop(), "cannot stop");
}

static char milter_getsymval__doc__[] =
"getsymval(String) -> String\n\
Returns a symbol's value.  Context-dependent, and unclear from the dox.";

static PyObject *
milter_getsymval(PyObject *self, PyObject *args)
{
   char *str;

   if (!PyArg_ParseTuple(args, "s", &str))
      return NULL;
   /* Error check here? */
   return Py_BuildValue("s", smfi_getsymval(context, str));
}

static char milter_setreply__doc__[] =
"setreply(rcode, xcode, message) -> SUCCESS or FAILURE\n\
Sets the specific reply code to be used in response to the active command.\
rcode - The three-digit (RFC 821) SMTP reply code to be returned\
xcode - The extended (RFC 2034) reply code\
message - The text part of the SMTP reply\
These should all be strings.";

static PyObject *
milter_setreply(PyObject *self, PyObject *args)
{
  char *rcode;
  char *xcode;
  char *message;

   if (!PyArg_ParseTuple(args, "sss", &rcode, &xcode, &message))
      return NULL;
   return _generic_return(smfi_setreply(context, rcode, xcode, message),
			  "cannot set reply");
}

static char milter_addheader__doc__[] =
"addheader(field, value) -> SUCCESS or FAILURE\n\
Add a header to the message. This header is not passed to other \
filters. It is not checked for standards compliance; the mail filter \
must ensure that no protocols are violated as a result of adding this \
header.\n\
field - header field name\
value - header field value\
Both are strings.  This function can only be called from the EOM callback.";

static PyObject *
milter_addheader(PyObject *self, PyObject *args)
{
  char *headerf;
  char *headerv;

   if (!PyArg_ParseTuple(args, "ss", &headerf, &headerv))
      return NULL;
   return _generic_return(smfi_addheader(context, headerf, headerv),
			  "cannot add header");
}

static char milter_chgheader__doc__[] =
"chgheader(field, int, value) -> SUCCESS or FAILURE\n\
Change/delete a header in the message. \
It is not checked for standards compliance; the mail filter \
must ensure that no protocols are violated as a result of adding this \
header.\n\
field - header field name\
int - the Nth occurence of this header\
value - header field value\
field and value are strings.  This function can only be called from the EOM \
callback.";

static PyObject *
milter_chgheader(PyObject *self, PyObject *args)
{
  char *headerf;
  int index;
  char *headerv;

   if (!PyArg_ParseTuple(args, "sis", &headerf, &index, &headerv))
      return NULL;
   return _generic_return(smfi_chgheader(context, headerf, index, headerv),
			  "cannot change header");
}

static char milter_addrcpt__doc__[] =
"addrcpt(string) -> SUCCESS or FAILURE\n\
Add a recipient to the envelope.  This function can only be called from the \
EOM callback.";

static PyObject *
milter_addrcpt(PyObject *self, PyObject *args)
{
  char *rcpt;

   if (!PyArg_ParseTuple(args, "s", &rcpt))
      return NULL;
   return _generic_return(smfi_addrcpt(context, rcpt), "cannot add recipient");
}

static char milter_delrcpt__doc__[] =
"delrcpt(string) -> SUCCESS or FAILURE\n\
Delete a recipient from the envelope.  This function can only be called from the \
EOM callback.";

static PyObject *
milter_delrcpt(PyObject *self, PyObject *args)
{
  char *rcpt;

   if (!PyArg_ParseTuple(args, "s", &rcpt))
      return NULL;
   return _generic_return(smfi_delrcpt(context, rcpt),
			  "cannot delete recipient");
}

static char milter_replacebody__doc__[] =
"replacebody(string) -> SUCCESS or FAILURE\n\
Replace the body of the message. This routine may be called multiple \
times if the body is longer than convenient to send in one call. End of \
line should be represented as Carriage-Return/Line Feed.  This function \
can only be called from the EOM callback.";

static PyObject *
milter_replacebody(PyObject *self, PyObject *args)
{
  char *bodyp;
  int bodylen;
  
  if (!PyArg_ParseTuple(args, "s#", &bodyp, &bodylen))
    return NULL;
  return _generic_return(smfi_replacebody(context, bodyp, bodylen),
			 "cannot replace message body");
}

static char milter_setpriv__doc__[] =
"setpriv(object) -> SUCCESS or FAILURE\n\
Associates any Python object with the current context.  Use this to \
provide thread-safe storage for data instead of using global variables \
for things like filenames, etc.  This function stores only one object \
per context, and cannot be called until there exists a message context; ie, \
it probably must be called from one of the event callbacks.";

static PyObject *
milter_setpriv(PyObject *self, PyObject *args)
{
  PyObject *o;
  
  if (!PyArg_ParseTuple(args, "O:setpriv", &o))
    return NULL;
  if (context == NULL) {
    PyErr_SetString(MilterError, "no context");
    return NULL;
  }
  /* PyArg_ParseTuple's O format does not increase the reference count on */
  /* the target.  Since we're going to save it and almost certainly assign */
  /* to another object later, we must assure that it doesn't get blasted */
  /* in the meantime.  Hence, we incref it here, and only decref it in */
  /* the stop method, since we don't know when getpriv will no longer be */
  /* called. */
  Py_INCREF(o);
  return _generic_return(smfi_setpriv(context, o), "cannot set private data");
}

static char milter_getpriv__doc__[] =
"getpriv() -> SUCCESS or FAILURE\n\
Returns the Python object associated with the current context (if any).  Use \
this in conjunction with setpriv to keep track of data in a thread-safe \
manner.  This function cannot be called until there exists a message \
context; ie, it probably must be called from one of the event callbacks.";

static PyObject *
milter_getpriv(PyObject *self, PyObject *args)
{
  PyObject *o;
  
  if (!PyArg_ParseTuple(args, ""))
    return NULL;
  if (context == NULL) {
    PyErr_SetString(MilterError, "no context");
    return NULL;
  }
  o = (PyObject *)smfi_getpriv(context);
  if (o == NULL) {
    PyErr_SetString(MilterError, "cannot get private data");
    return NULL;
  }
  /* Py_BuildValue does *not* incref the object - I don't *think* I have */
  /* to here, considering what happens in setpriv, but I'm not sure */
  return Py_BuildValue("O", o);
}

struct smfiDesc description =
{  /* Set some reasonable defaults */
  "pythonfilter",
  SMFI_VERSION,
  SMFI_CURR_ACTS,
  milter_wrap_connect,
  milter_wrap_helo,
  milter_wrap_envfrom,
  milter_wrap_envrcpt,
  milter_wrap_header,
  milter_wrap_eoh,
  milter_wrap_body,
  milter_wrap_eom,
  milter_wrap_abort,
  milter_wrap_close
};

static PyMethodDef milter_methods[] = {
   { "set_name",             milter_set_name,             METH_VARARGS, milter_set_name__doc__},
   { "set_flags",            milter_set_flags,            METH_VARARGS, milter_set_flags__doc__},
   { "set_connect_callback", milter_set_connect_callback, METH_VARARGS, milter_set_connect_callback__doc__},
   { "set_helo_callback",    milter_set_helo_callback,    METH_VARARGS, milter_set_helo_callback__doc__},
   { "set_envfrom_callback", milter_set_envfrom_callback, METH_VARARGS, milter_set_envfrom_callback__doc__},
   { "set_envrcpt_callback", milter_set_envrcpt_callback, METH_VARARGS, milter_set_envrcpt_callback__doc__},
   { "set_header_callback",  milter_set_header_callback,  METH_VARARGS, milter_set_header_callback__doc__},
   { "set_eoh_callback",     milter_set_eoh_callback,     METH_VARARGS, milter_set_eoh_callback__doc__},
   { "set_body_callback",    milter_set_body_callback,    METH_VARARGS, milter_set_body_callback__doc__},
   { "set_eom_callback",     milter_set_eom_callback,     METH_VARARGS, milter_set_eom_callback__doc__},
   { "set_abort_callback",   milter_set_abort_callback,   METH_VARARGS, milter_set_abort_callback__doc__},
   { "set_close_callback",   milter_set_close_callback,   METH_VARARGS, milter_set_close_callback__doc__},
   { "register",             milter_register,             METH_VARARGS, milter_register__doc__},
   { "main",                 milter_main,                 METH_VARARGS, milter_main__doc__},
   { "setdbg",               milter_setdbg,               METH_VARARGS, milter_setdbg__doc__},
   { "settimeout",           milter_settimeout,           METH_VARARGS, milter_settimeout__doc__},
   { "setconn",              milter_setconn,              METH_VARARGS, milter_setconn__doc__},
   { "stop",                 milter_stop,                 METH_VARARGS, milter_stop__doc__},
   { "getsymval",            milter_getsymval,            METH_VARARGS, milter_getsymval__doc__},
   { "setreply",             milter_setreply,             METH_VARARGS, milter_setreply__doc__},
   { "addheader",            milter_addheader,            METH_VARARGS, milter_addheader__doc__},
   { "chgheader",            milter_chgheader,            METH_VARARGS, milter_chgheader__doc__},
   { "addrcpt",              milter_addrcpt,              METH_VARARGS, milter_addrcpt__doc__},
   { "delrcpt",              milter_delrcpt,              METH_VARARGS, milter_delrcpt__doc__},
   { "replacebody",          milter_replacebody,          METH_VARARGS, milter_replacebody__doc__},
   { "setpriv",              milter_setpriv,              METH_VARARGS, milter_setpriv__doc__},
   { "getpriv",              milter_getpriv,              METH_VARARGS, milter_getpriv__doc__},
   { NULL, NULL }
};

static char milter_documentation[] =
"This module interfaces with Sendmail's libmilter functionality, allowing one"
"to write email filters directly in Python.\n"
"Libmilter is currently marked FFR, and needs to be explicitly installed. See"
"<sendmailsource>/libmilter/README for details on setting it up.\n"
;

void
initmilter(void)
{
   PyObject *m, *d;

   m = Py_InitModule4("milter", milter_methods, milter_documentation,
		      (PyObject*)NULL, PYTHON_API_VERSION);
   d = PyModule_GetDict(m);
   MilterError = PyErr_NewException("milter.error", NULL, NULL);
   PyDict_SetItemString(d,"error", MilterError);
   PyDict_SetItemString(d,"SUCCESS", PyInt_FromLong((long) MI_SUCCESS));
   PyDict_SetItemString(d,"FAILURE", PyInt_FromLong((long) MI_FAILURE));
   PyDict_SetItemString(d,"VERSION", PyInt_FromLong((long) SMFI_VERSION));
   PyDict_SetItemString(d,"ADDHDRS", PyInt_FromLong((long) SMFIF_ADDHDRS));
   PyDict_SetItemString(d,"CHGBODY", PyInt_FromLong((long) SMFIF_CHGBODY));
   PyDict_SetItemString(d,"MODBODY", PyInt_FromLong((long) SMFIF_MODBODY));
   PyDict_SetItemString(d,"ADDRCPT", PyInt_FromLong((long) SMFIF_ADDRCPT));
   PyDict_SetItemString(d,"DELRCPT", PyInt_FromLong((long) SMFIF_DELRCPT));
   PyDict_SetItemString(d,"CHGHDRS", PyInt_FromLong((long) SMFIF_CHGHDRS));
   PyDict_SetItemString(d,"V1_ACTS", PyInt_FromLong((long) SMFI_V1_ACTS));
   PyDict_SetItemString(d,"V2_ACTS", PyInt_FromLong((long) SMFI_V2_ACTS));
   PyDict_SetItemString(d,"CURR_ACTS", PyInt_FromLong((long) SMFI_CURR_ACTS));
   PyDict_SetItemString(d,"CONTINUE", PyInt_FromLong((long) SMFIS_CONTINUE));
   PyDict_SetItemString(d,"REJECT", PyInt_FromLong((long) SMFIS_REJECT));
   PyDict_SetItemString(d,"DISCARD", PyInt_FromLong((long) SMFIS_DISCARD));
   PyDict_SetItemString(d,"ACCEPT", PyInt_FromLong((long) SMFIS_ACCEPT));
   PyDict_SetItemString(d,"TEMPFAIL", PyInt_FromLong((long) SMFIS_TEMPFAIL));
}
